@@ -1,6 +1,6 @@
 from guillotina.commands import Command
 from guillotina.commands.utils import change_transaction_strategy
-from guillotina.component import getUtility
+from guillotina.component import get_utility
 from guillotina.interfaces import ICatalogUtility
 from guillotina.utils import get_containers
 from guillotina_elasticsearch.reindex import Reindexer
@@ -34,7 +34,7 @@ class ReindexCommand(Command):
         return parser
 
     async def reindex_all(self, arguments):
-        search = getUtility(ICatalogUtility)
+        search = get_utility(ICatalogUtility)
         await asyncio.sleep(1)  # since something initialize custom types...
         async for _, tm, container in get_containers(self.request):
             try:
@@ -43,7 +43,8 @@ class ReindexCommand(Command):
                     log_details=arguments.log_details,
                     memory_tracking=arguments.memory_tracking,
                     reindex_security=arguments.reindex_security,
-                    mapping_only=arguments.mapping_only)
+                    mapping_only=arguments.mapping_only,
+                    cache=False)
                 await self.reindexer.reindex(container)
                 seconds = int(time.time() - self.reindexer.start_time)
                 logger.warning(f'''Finished reindex:
