@@ -1,9 +1,11 @@
 from guillotina import directives
 from guillotina.catalog.catalog import DefaultCatalogDataAdapter
+from guillotina.component import get_adapter
 from guillotina.component import get_utilities_for
 from guillotina.content import iter_schemata_for_type
 from guillotina.db.cache.dummy import DummyCache
 from guillotina.directives import merged_tagged_value_dict
+from guillotina.event import notify
 from guillotina.exceptions import NoIndexField
 from guillotina.interfaces import IAsyncBehavior
 from guillotina.interfaces import ICatalogDataAdapter
@@ -16,13 +18,19 @@ from guillotina.transactions import managed_transaction
 from guillotina.utils import apply_coroutine
 from guillotina.utils import get_content_path
 from guillotina.utils import get_current_request
-from guillotina_elasticsearch.utils import noop_response
-from guillotina.event import notify
 from guillotina_elasticsearch.events import IndexProgress
+from guillotina_elasticsearch.interfaces import DOC_TYPE
+from guillotina_elasticsearch.interfaces import IIndexActive
+from guillotina_elasticsearch.interfaces import IIndexManager
+from guillotina_elasticsearch.utils import find_index_manager
+from guillotina_elasticsearch.utils import get_migration_lock
+from guillotina_elasticsearch.utils import noop_response
 from os.path import join
 
-import aioes
+import aioelasticsearch
 import asyncio
+import backoff
+import elasticsearch.exceptions
 import gc
 import json
 import logging
@@ -30,13 +38,6 @@ import resource
 import time
 
 
-try:
-    from guillotina.utils import clear_conn_statement_cache
-except ImportError:
-    def clear_conn_statement_cache(conn):
-        pass
-
-
 logger = logging.getLogger('guillotina_elasticsearch')
 
 
@@ -125,18 +126,14 @@ class Migrator:
     10. Refresh db container ob
     11. Point alias at next index
     12. Delete old index
-
-
-    TODO:
-        - optionally fill metadata in indexing
-            - requires more work...
     '''
 
     def __init__(self, utility, context, response=noop_response, force=False,
                  log_details=False, memory_tracking=False, request=None,
-                 bulk_size=40, full=False, reindex_security=False, mapping_only=False):
+                 bulk_size=40, full=False, reindex_security=False,
+                 mapping_only=False, index_manager=None, children_only=False,
+                 lookup_index=False, cache=True):
         self.utility = utility
-        self.conn = utility.conn
         self.context = context
         self.response = response
         self.force = force
@@ -145,17 +142,30 @@ class Migrator:
         self.memory_tracking = memory_tracking
         self.bulk_size = bulk_size
         self.reindex_security = reindex_security
+        self.children_only = children_only
+        self.lookup_index = lookup_index
         if mapping_only and full:
-            raise Exception('Can not do a full reindex and a mapping only migration')
+            raise Exception(
+                'Can not do a full reindex and a mapping only migration')
         self.mapping_only = mapping_only
 
         if request is None:
             self.request = get_current_request()
         else:
             self.request = request
-        # make sure that we don't cache requests...
-        self.request._txn._cache = DummyCache(self.request._txn)
+
+        if not cache:
+            # make sure that we don't cache requests...
+            self.request._txn._cache = DummyCache(self.request._txn)
+
         self.container = self.request.container
+        self.conn = utility.get_connection()
+
+        if index_manager is None:
+            self.index_manager = get_adapter(self.container, IIndexManager)
+        else:
+            self.index_manager = index_manager
+
         self.interaction = IInteraction(self.request)
         self.indexer = Indexer()
 
@@ -175,35 +185,36 @@ class Migrator:
         self.copied_docs = 0
 
         self.work_index_name = None
+        self.sub_indexes = []
 
     def per_sec(self):
         return self.processed / (time.time() - self.index_start_time)
 
     async def create_next_index(self):
-        version = await self.utility.get_version(self.container,
-                                                 request=self.request)
-        next_version = version + 1
-        index_name = await self.utility.get_index_name(self.container,
-                                                       request=self.request)
-        next_index_name = index_name + '_' + str(next_version)
+        async with managed_transaction(
+                self.request, write=True, adopt_parent_txn=True) as txn:
+            await txn.refresh(await self.index_manager.get_registry())
+            next_index_name = await self.index_manager.start_migration()
         if await self.conn.indices.exists(next_index_name):
             if self.force:
                 # delete and recreate
                 self.response.write('Clearing index')
                 resp = await self.conn.indices.delete(next_index_name)
                 assert resp['acknowledged']
-        await self.conn.indices.create(next_index_name)
-        return next_version, next_index_name
+        await self.utility.create_index(next_index_name, self.index_manager)
+        return next_index_name
 
     async def copy_to_next_index(self):
         conn_es = await self.conn.transport.get_connection()
-        real_index_name = await self.utility.get_index_name(self.container,
-                                                            self.request)
-        async with conn_es._session.post(
-                str(conn_es._base_url) + '_reindex',
+        real_index_name = await self.index_manager.get_index_name()
+        async with conn_es.session.post(
+                join(str(conn_es.base_url), '_reindex'),
                 params={
                     'wait_for_completion': 'false'
                 },
+                headers={
+                    'Content-Type': 'application/json'
+                },
                 data=json.dumps({
                     "source": {
                         "index": real_index_name,
@@ -217,16 +228,20 @@ class Migrator:
             self.active_task_id = task_id = data['task']
             while True:
                 await asyncio.sleep(10)
-                async with conn_es._session.get(
-                        str(conn_es._base_url) + '_tasks/' + task_id) as resp:
+                async with conn_es.session.get(
+                        join(str(conn_es.base_url), '_tasks', task_id),
+                        headers={
+                            'Content-Type': 'application/json'
+                        }) as resp:
                     if resp.status in (400, 404):
                         break
                     data = await resp.json()
                     if data['completed']:
                         break
                     status = data["task"]["status"]
-                    self.response.write(f'{status["created"]}/{status["total"]} - '
-                                        f'Copying data to new index. task id: {task_id}')
+                    self.response.write(
+                        f'{status["created"]}/{status["total"]} - '
+                        f'Copying data to new index. task id: {task_id}')
                     self.copied_docs = status["created"]
 
             self.active_task_id = None
@@ -235,27 +250,30 @@ class Migrator:
             if len(failures) > 0:
                 failures = json.dumps(failures, sort_keys=True, indent=4,
                                       separators=(',', ': '))
-                self.response.write(f'Reindex encountered failures: {failures}')
+                self.response.write(
+                    f'Reindex encountered failures: {failures}')
             else:
-                self.response.write(f'Finished copying to new index: {self.copied_docs}')
+                self.response.write(
+                    f'Finished copying to new index: {self.copied_docs}')
 
     async def get_all_uids(self):
         self.response.write('Retrieving existing doc ids')
         page_size = 3000
         ids = []
-        index_name = await self.utility.get_index_name(self.container)
+        index_name = await self.index_manager.get_index_name()
         result = await self.conn.search(
             index=index_name,
             scroll='2m',
             size=page_size,
             stored_fields='',
+            _source=False,
             body={
                 "sort": ["_doc"]
             })
         ids.extend([r['_id'] for r in result['hits']['hits']])
         scroll_id = result['_scroll_id']
         while scroll_id:
-            result = await self.utility.conn.scroll(
+            result = await self.conn.scroll(
                 scroll_id=scroll_id,
                 scroll='2m'
             )
@@ -264,7 +282,8 @@ class Migrator:
             ids.extend([r['_id'] for r in result['hits']['hits']])
             self.response.write(f'Retrieved {len(ids)} doc ids')
             scroll_id = result['_scroll_id']
-        self.response.write(f'Retrieved {len(ids)}. Copied {self.copied_docs} docs')
+        self.response.write(
+            f'Retrieved {len(ids)}. Copied {self.copied_docs} docs')
         return ids
 
     async def calculate_mapping_diff(self):
@@ -272,71 +291,35 @@ class Migrator:
         all we care about is new fields...
         Missing ones are ignored and we don't care about it.
         '''
-        diffs = {}
-        existing_index_name = await self.utility.get_real_index_name(
-            self.container, self.request)
-        existing_mappings = await self.conn.indices.get_mapping(existing_index_name)
-        existing_mappings = existing_mappings[existing_index_name]['mappings']
-        next_mappings = await self.conn.indices.get_mapping(self.work_index_name)
+        next_mappings = await self.conn.indices.get_mapping(
+            self.work_index_name)
         next_mappings = next_mappings[self.work_index_name]['mappings']
+        next_mappings = next_mappings[DOC_TYPE]['properties']
 
-        changes = False
-        for type_name in existing_mappings.keys():
-            if type_name not in next_mappings:
-                # copy over orphaned type otherwise move will potentially not work
-                # any orphaned doc types will need to be manually deleted for now...
-                mapping = existing_mappings[type_name]
-                properties = mapping['properties']
-                # need to make sure to normalize field definitions so they are inline
-                # with new mappings otherwise you could get conflicting definitions
-                for field_name in properties.keys():
-                    for check_type_name in next_mappings.keys():
-                        if field_name in next_mappings[check_type_name]['properties']:
-                            properties[field_name] = next_mappings[
-                                check_type_name]['properties'][field_name]
-                            break
-                # and install new mapping
-                await self.utility.conn.indices.put_mapping(
-                    self.work_index_name, type_name, mapping)
-                changes = True
-        if changes:
-            # we add to the mappings so we need to update...
-            next_mappings = await self.conn.indices.get_mapping(self.work_index_name)
-            next_mappings = next_mappings[self.work_index_name]['mappings']
+        existing_index_name = await self.index_manager.get_real_index_name()
+        try:
+            existing_mappings = await self.conn.indices.get_mapping(
+                existing_index_name)
+        except elasticsearch.exceptions.NotFoundError:
+            # allows us to upgrade when no index is present yet
+            return next_mappings
 
-        for type_name, schema in get_utilities_for(IResourceFactory):
-            new_definitions = {}
-            if type_name not in existing_mappings:
-                diffs[type_name] = next_mappings[type_name]['properties']
-                continue
+        existing_mappings = existing_mappings[existing_index_name]['mappings']
+        existing_mappings = existing_mappings[DOC_TYPE]['properties']
 
-            existing_mapping = existing_mappings[type_name]['properties']
-            next_mapping = next_mappings[type_name]['properties']
-
-            for field_name, definition in next_mapping.items():
-                definition = _clean_mapping(definition)
-                if (field_name not in existing_mapping or
-                        definition != _clean_mapping(existing_mapping[field_name])):
-                    new_definitions[field_name] = definition
-            if len(new_definitions) > 0:
-                diffs[type_name] = new_definitions
-
-        for type_name, mapping in existing_mappings.items():
-            if type_name not in next_mappings:
-                # special case here... we need to import this mapping still
-                # in order for the index copy to work correctly if docs ref it
-                self.response.write(f'Backporting mapping of {type_name} to new '
-                                    f'even though it is not defined anymore')
-                await self.conn.indices.put_mapping(
-                    self.work_index_name, type_name, mapping)
-
-        return diffs
+        new_definitions = {}
+        for field_name, definition in next_mappings.items():
+            definition = _clean_mapping(definition)
+            if (field_name not in existing_mappings or
+                    definition != _clean_mapping(existing_mappings[field_name])):  # noqa
+                new_definitions[field_name] = definition
+        return new_definitions
 
     async def process_folder(self, ob):
         for key in await ob.async_keys():
             try:
                 item = await ob._p_jar.get_child(ob, key)
-            except KeyError:
+            except (KeyError, ModuleNotFoundError):
                 continue
             if item is None:
                 continue
@@ -356,7 +339,6 @@ class Migrator:
             - else, do nothing
             - remove for list of existing doc ids
         '''
-        clear_conn_statement_cache(await ob._p_jar.get_connection())
         full = False
         if ob.uuid not in self.existing:
             self.missing.append(ob.uuid)
@@ -366,12 +348,18 @@ class Migrator:
         await self.index_object(ob, full=full)
         self.processed += 1
 
-        if IFolder.providedBy(ob):
-            await self.process_folder(ob)
+        if IIndexActive.providedBy(ob):
+            self.sub_indexes.append(ob)
+        else:
+            if IFolder.providedBy(ob):
+                await self.process_folder(ob)
 
-        if not IContainer.providedBy(ob):
-            del ob.__annotations__
-        del ob
+            if not IContainer.providedBy(ob):
+                try:
+                    del self.container.__gannotations__
+                except AttributeError:
+                    del self.container.__annotations__
+            del ob
 
     async def index_object(self, ob, full=False):
         batch_type = 'update'
@@ -379,41 +367,43 @@ class Migrator:
             try:
                 data = ISecurityInfo(ob)()
             except TypeError:
-                self.response.write(f'could not index {ob}')
+                self.response.write(f'Could not index {ob}')
                 return
         elif full or self.full:
             try:
                 data = await ICatalogDataAdapter(ob)()
             except TypeError:
-                self.response.write(f'could not index {ob}')
+                self.response.write(f'Could not index {ob}')
                 return
             batch_type = 'index'
         else:
-            if ob.type_name not in self.mapping_diff:
-                # no fields change, ignore this guy...
-                if self.log_details:
-                    self.response.write(f'({self.processed} {int(self.per_sec())}) '
-                                        f'(skipped) Object: {get_content_path(ob)}, '
-                                        f'Type: {batch_type}, Buffer: {len(self.batch)}')
-                return
             data = {
-                'type_name': ob.type_name  # always need this one...
+                # always need these...
+                'type_name': ob.type_name
             }
-            for index_name in self.mapping_diff[ob.type_name].keys():
+            for index_name in self.mapping_diff.keys():
                 val = await self.indexer.get_value(ob, index_name)
                 if val is not None:
                     data[index_name] = val
 
+        if ob._p_serial:
+            data['tid'] = ob._p_serial
         self.indexed += 1
         self.batch[ob.uuid] = {
             'action': batch_type,
             'data': data
         }
 
+        if self.lookup_index:
+            im = find_index_manager(ob)
+            if im:
+                self.batch[ob.uuid]['__index__'] = await im.get_index_name()
+
         if self.log_details:
-            self.response.write(f'({self.processed} {int(self.per_sec())}) '
-                                f'Object: {get_content_path(ob)}, '
-                                f'Type: {batch_type}, Buffer: {len(self.batch)}')
+            self.response.write(
+                f'({self.processed} {int(self.per_sec())}) '
+                f'Object: {get_content_path(ob)}, '
+                f'Type: {batch_type}, Buffer: {len(self.batch)}')
 
         await self.attempt_flush()
 
@@ -425,10 +415,11 @@ class Migrator:
             gc.collect()
             if self.memory_tracking:
                 total_memory = round(
-                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.0, 1)
-                self.response.write(b'Memory usage: % 2.2f MB, cleaned: %d, total in-memory obs: %d' % (
-                    total_memory, num, len(gc.get_objects())))
-            self.response.write(b'Indexing new batch, totals: (%d %d/sec)\n' % (
+                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.0, 1)  # noqa
+                self.response.write(
+                    b'Memory usage: % 2.2f MB, cleaned: %d, total in-memory obs: %d' % (  # noqa
+                        total_memory, num, len(gc.get_objects())))
+            self.response.write(b'Indexing new batch, totals: (%d %d/sec)\n' % (  # noqa
                 self.indexed, int(self.per_sec()),
             ))
         if len(self.batch) >= self.bulk_size:
@@ -444,13 +435,16 @@ class Migrator:
                 await asyncio.wait_for(future, None)
         self.reindex_futures = []
 
+    @backoff.on_exception(
+        backoff.constant,
+        (asyncio.TimeoutError, elasticsearch.exceptions.ConnectionTimeout),
+        interval=1, max_tries=5)
     async def _index_batch(self, batch):
         bulk_data = []
         for _id, payload in batch.items():
-            doc_type = payload['data']['type_name']
+            index = payload.pop('__index__', self.work_index_name)
             action_data = {
-                '_index': self.work_index_name,
-                '_type': doc_type,
+                '_index': index,
                 '_id': _id
             }
             data = payload['data']
@@ -462,8 +456,8 @@ class Migrator:
             })
             if payload['action'] != 'delete':
                 bulk_data.append(data)
-        results = await self.utility.conn.bulk(
-            index=self.work_index_name, doc_type=None,
+        results = await self.conn.bulk(
+            index=self.work_index_name, doc_type=DOC_TYPE,
             body=bulk_data)
         if results['errors']:
             errors = []
@@ -473,12 +467,17 @@ class Migrator:
                         continue
                     if 'status' in value and value['status'] != 200:
                         _id = value.get('_id')
-                        errors.append(f'{_id}: {value["status"]}')
 
-                        # retry conflict errors and thread pool capacity limits
+                        # retry conflict errors and thread pool exceeded errors
                         if value['status'] in (409, 429):
                             self.batch[_id] = batch[_id]
-            logger.warning(f'Error bulk putting: {results}')
+                        elif value['status'] == 404:
+                            self.batch[_id] = batch[_id]
+                            self.batch[_id]['action'] = 'index'
+                        else:
+                            errors.append(f'{_id}: {value["status"]}')
+            if len(errors) > 0:
+                logger.warning(f'Error bulk putting: {errors}')
 
     async def flush(self):
         if len(self.batch) == 0:
@@ -504,72 +503,57 @@ class Migrator:
             except KeyError:
                 ob = None
             if ob is None:
-                # this is dumb... since we don't have the doc type on us, we
-                # need to ask elasticsearch for it again...
-                # elasticsearch does not allow deleting without the doc type
-                # even though you can query for a doc without it... argh
                 try:
-                    doc = await self.conn.get(self.work_index_name, uuid,
-                                              _source=False)
                     self.batch[uuid] = {
                         'action': 'delete',
-                        'data': {'type_name': doc['_type']}
+                        'data': {}
                     }
                     await self.attempt_flush()
                     # no longer present on db, this was orphaned
                     self.orphaned.append(uuid)
-                except aioes.exception.NotFoundError:
+                except aioelasticsearch.exceptions.NotFoundError:
                     # it was deleted in the meantime so we're actually okay
-                    pass
+                    self.orphaned.append(uuid)
             else:
-                # XXX this should not happen so log it. Maybe we'll try doing something
-                # about it another time...
+                # XXX this should not happen so log it. Maybe we'll try
+                # doing something about it another time...
                 self.errors.append({
                     'type': 'unprocessed',
                     'uuid': uuid
                 })
-                # we re-query es to get full path of ob
-                # doc = self.utility.conn.get(
-                #     await self.utility.get_index_name(), fields='path'
-                # )
-                # ob = await traverse(self.request, self.container,
-                #                        doc['_source']['path'].strip('/').split('/'))
-                # await self.index_object(ob, full=True)
 
     async def setup_next_index(self):
         self.response.write(b'Creating new index')
-        async with managed_transaction(self.request, write=True, adopt_parent_txn=True):
-            await self.utility.disable_next_index(self.context, request=self.request)
-
-        async with managed_transaction(self.request, write=True, adopt_parent_txn=True):
-            self.next_index_version, self.work_index_name = await self.create_next_index()
-            await self.utility.install_mappings_on_index(self.work_index_name)
-            await self.utility.activate_next_index(
-                self.container, self.next_index_version, request=self.request,
-                force=self.force)
+        async with get_migration_lock(
+                await self.index_manager.get_index_name()):
+            self.work_index_name = await self.create_next_index()
+            return self.work_index_name
 
     async def cancel_migration(self):
         # canceling the migration, clearing index
         self.response.write('Canceling migration')
-        async with managed_transaction(self.request, write=True, adopt_parent_txn=True):
-            await self.utility.disable_next_index(self.context, request=self.request)
+        async with managed_transaction(
+                self.request, write=True, adopt_parent_txn=True):
+            await self.index_manager.cancel_migration()
             self.response.write('Next index disabled')
         if self.active_task_id is not None:
             self.response.write('Canceling copy of index task')
             conn_es = await self.conn.transport.get_connection()
-            async with conn_es._session.post(
-                    str(conn_es._base_url) + '_tasks/' + self.active_task_id + '/_cancel'):
-                asyncio.sleep(5)
+            async with conn_es.session.post(
+                    join(str(conn_es.base_url),
+                         '_tasks', self.active_task_id, '_cancel'),
+                    headers={
+                        'Content-Type': 'application/json'
+                    }):
+                await asyncio.sleep(5)
         if self.work_index_name:
             self.response.write('Deleting new index')
             await self.conn.indices.delete(self.work_index_name)
         self.response.write('Migration canceled')
 
     async def run_migration(self):
-        alias_index_name = await self.utility.get_index_name(self.container,
-                                                             request=self.request)
-        existing_index = await self.utility.get_real_index_name(self.container,
-                                                                request=self.request)
+        alias_index_name = await self.index_manager.get_index_name()
+        existing_index = await self.index_manager.get_real_index_name()
 
         await self.setup_next_index()
 
@@ -579,44 +563,86 @@ class Migrator:
         self.response.write(f'Caculated mapping diff: {diff}')
 
         if not self.full:
-            # if full, we're reindexing everything does not matter what anyways, so skip
-            self.response.write('Copying initial index data from existing index into new')
-            await self.copy_to_next_index()
-            self.response.write('Copying initial index data finished')
-
+            # if full, we're reindexing everything does not matter what
+            # anyways, so skip
+            self.response.write(f'Copying initial index {existing_index} '
+                                f'into {self.work_index_name}')
+            try:
+                await self.copy_to_next_index()
+                self.response.write('Copying initial index data finished')
+            except elasticsearch.exceptions.NotFoundError:
+                self.response.write('No initial index to copy to')
         if not self.mapping_only:
-            self.existing = await self.get_all_uids()
+            try:
+                self.existing = await self.get_all_uids()
+            except elasticsearch.exceptions.NotFoundError:
+                pass
 
             self.index_start_time = time.time()
-            await self.process_object(self.context)  # this is recursive
+            if self.children_only or IContainer.providedBy(self.context):
+                await self.process_folder(self.context)  # this is recursive
+            else:
+                await self.process_object(self.context)  # this is recursive
 
             await self.check_existing()
 
             await self.flush()
             await self.join_futures()
 
-        async with self.utility._migration_lock:
+        async with get_migration_lock(
+                await self.index_manager.get_index_name()):
             self.response.write('Activating new index')
-            async with managed_transaction(self.request, write=True, adopt_parent_txn=True):
-                await self.utility.apply_next_index(self.container, self.request)
+            async with managed_transaction(
+                    self.request, write=True, adopt_parent_txn=True):
+                await self.index_manager.finish_migration()
             self.status = 'done'
 
             self.response.write(f'''Update alias({alias_index_name}):
 {existing_index} -> {self.work_index_name}
 ''')
-            await self.conn.indices.update_aliases({
-                "actions": [
-                    {"remove": {
-                        "alias": alias_index_name,
-                        "index": existing_index
-                    }},
-                    {"add": {
-                        "alias": alias_index_name,
-                        "index": self.work_index_name
-                    }}
-                ]
-            })
 
-        self.response.write('Delete old index')
-        await self.conn.indices.close(existing_index)
-        await self.conn.indices.delete(existing_index)
+            try:
+                await self.conn.indices.update_aliases({
+                    "actions": [
+                        {"remove": {
+                            "alias": alias_index_name,
+                            "index": existing_index
+                        }},
+                        {"add": {
+                            "alias": alias_index_name,
+                            "index": self.work_index_name
+                        }}
+                    ]
+                })
+            except elasticsearch.exceptions.NotFoundError:
+                await self.conn.indices.update_aliases({
+                    "actions": [
+                        {"add": {
+                            "alias": alias_index_name,
+                            "index": self.work_index_name
+                        }}
+                    ]
+                })
+
+        try:
+            await self.conn.indices.close(existing_index)
+            await self.conn.indices.delete(existing_index)
+            self.response.write('Old index deleted')
+        except elasticsearch.exceptions.NotFoundError:
+            pass
+
+        if len(self.sub_indexes) > 0:
+            self.response.write(
+                f'Migrating sub indexes: {len(self.sub_indexes)}')
+            for ob in self.sub_indexes:
+                im = get_adapter(ob, IIndexManager)
+                migrator = Migrator(
+                    self.utility, ob, response=self.response, force=self.force,
+                    log_details=self.log_details,
+                    memory_tracking=self.memory_tracking, request=self.request,
+                    bulk_size=self.bulk_size, full=self.full,
+                    reindex_security=self.reindex_security,
+                    mapping_only=self.mapping_only,
+                    index_manager=im, children_only=True)
+                self.response.write(f'Migrating index for: {ob}')
+                await migrator.run_migration()
